/*
 *    Copyright 2006 Intel Corporation
 * 
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 * 
 *        http://www.apache.org/licenses/LICENSE-2.0
 * 
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

#ifdef HAVE_CONFIG_H
#  include <dtn-config.h>
#endif

#include "BPQBlock.h"
#include "Bundle.h"
#include "BundleProtocol.h"
#include "SDNV.h"

namespace dtn {

// Setup our logging information
static const char* LOG = "/dtn/bundle/extblock/bpq";

BPQBlock::BPQBlock(Bundle* bundle)
{
    log_info_p(LOG, "BPQBlock::constructor()");

    if( bundle->recv_blocks().
        has_block(BundleProtocol::QUERY_EXTENSION_BLOCK) ) {

        log_debug_p(LOG, "BPQBlock found in Recv Block Vec => created remotly");
        initialise( const_cast<BlockInfo*> (bundle->recv_blocks().
                    find_block(BundleProtocol::QUERY_EXTENSION_BLOCK)) );

    } else if( bundle->api_blocks()->
               has_block(BundleProtocol::QUERY_EXTENSION_BLOCK) ) {

        log_debug_p(LOG, "BPQBlock found in API Block Vec => created locally");
        initialise( const_cast<BlockInfo*> (bundle->api_blocks()->
                    find_block(BundleProtocol::QUERY_EXTENSION_BLOCK)) );

    } else {
        log_err_p(LOG, "BPQ Block not found in bundle");
    }

    log_info_p(LOG, "leaving constructor");
}

//----------------------------------------------------------------------
BPQBlock::BPQBlock(BlockInfo* block)
{
    log_info_p(LOG, "BPQBlock::constructor()");

    ASSERT( block->type() == BundleProtocol::QUERY_EXTENSION_BLOCK);

    initialise(block);

    log_info_p(LOG, "leaving constructor");
}

//----------------------------------------------------------------------
BPQBlock::~BPQBlock()
{
    log_info_p(LOG, "BPQBlock: destructor");
    if ( query_val_ != NULL ){
        free(query_val_);
        query_val_ = NULL;
    }
}

//----------------------------------------------------------------------
int
BPQBlock::write_to_buffer(u_char* buf, size_t len)
{
    int encoding_len=0; 
    u_int i=0, j=0;

    // BPQ-kind             1-byte
    if ( i < len ) 
        buf[i++] = (u_char) kind_;
    else
        return -1;

    // matching rule type   1-byte
    if ( i < len )
        buf[i++] = (u_char) matching_rule_;
    else
        return -1;

    // query-length         SDNV
    if ( i < len &&
        (encoding_len = SDNV::encode (query_len_, &(buf[i]), len -i)) >= 0 ) {
        i += encoding_len;
    } else {
        log_err_p(LOG, "Error encoding _BPQ query length");
        return -1;
    }

    // query-value           n-bytes
    for (j=0; query_val_ != NULL && i < len && j < query_len_; i++, j++)
        buf[i] = query_val_[j];    

    // fragment-length		SDNV
    if ( i < len &&
        (encoding_len = SDNV::encode (frag_len(), &(buf[i]), len -i)) >= 0 ) {
        i += encoding_len;
    } else {
        log_err_p(LOG, "Error encoding _BPQ fragment length");
        return -1;
    }

    // fragment-values		SDNV
    BPQFragmentVec::const_iterator iter;
    for (iter  = fragments_.begin();
    	 iter != fragments_.end();
    	 ++iter) {

        if ( i < len &&
            (encoding_len = SDNV::encode (iter->offset(), &(buf[i]), len -i)) >= 0 ) {
            i += encoding_len;
        } else {
            log_err_p(LOG, "Error encoding _BPQ individual fragment offset");
            return -1;
        }

        if ( i < len &&
            (encoding_len = SDNV::encode (iter->length(), &(buf[i]), len -i)) >= 0 ) {
            i += encoding_len;
        } else {
            log_err_p(LOG, "Error encoding _BPQ individual fragment length");
            return -1;
        }
    }

    ASSERT ( i == this->length())

    return i;
}

//----------------------------------------------------------------------
u_int
BPQBlock::length() const
{
    // initial size {kind, matching rule}
    u_int len = 2; 
        
    len += SDNV::encoding_len(query_len_);
    len += query_len_;
    len += SDNV::encoding_len(frag_len());

    BPQFragmentVec::const_iterator iter;
	for (iter  = fragments_.begin();
		 iter != fragments_.end();
		 ++iter) {

		len += SDNV::encoding_len(iter->offset());
		len += SDNV::encoding_len(iter->length());
	}

    return len;
}

//----------------------------------------------------------------------
bool
BPQBlock::match(const BPQBlock* other) const
{
    return query_len_ == other->query_len() &&
           strncmp( (char*)query_val_, (char*)other->query_val(),
                     query_len_ ) == 0;
}

//----------------------------------------------------------------------
int
BPQBlock::initialise(BlockInfo* b)
{
    ASSERT ( b != NULL);

    int decoding_len=0;
	u_int i=0, j=0, offset=0, length=0, full_len=0;
	u_int frag_count=0, frag_off=0, frag_len=0;
	u_char* buf = 0;
    BlockInfo* block = b;

    /**************************************************************************
     * Begin extracting block length with lots of logging
     *************************************************************************/
    log_debug_p(LOG, "block: data_length() = %d", block->data_length());
    log_debug_p(LOG, "block: data_offset() = %d", block->data_offset());
    log_debug_p(LOG, "block: full_length() = %d", block->full_length());
    log_debug_p(LOG, "block: complete() = %s",
        (block->complete()) ? "true" : "false" );

    log_debug_p(LOG, "block: reloaded() = %s",
        (block->reloaded()) ? "true" : "false" );

    if ( b->source() != NULL ) {
        BlockInfo* block_src = const_cast<BlockInfo*>(b->source());;

        log_debug_p(LOG, "block_src: data_length() = %d", block_src->data_length());
        log_debug_p(LOG, "block_src: data_offset() = %d", block_src->data_offset());
        log_debug_p(LOG, "block_src: full_length() = %d", block_src->full_length());
        log_debug_p(LOG, "block_src: complete() = %s",
            (block_src->complete()) ? "true" : "false" );
    
        log_debug_p(LOG, "block_src: reloaded() = %s",
            (block_src->reloaded()) ? "true" : "false" );
    }


    offset = block->data_offset();
    length = block->data_length();
    full_len = block->full_length();

    if ( full_len != offset + length ) {
        log_err_p(LOG, "BPQBlock::initialise: full_len != offset + length");
    }

    if ( block->writable_contents()->buf_len() < full_len ){
        log_err_p(LOG, "BPQBlock::initialise:  buf_len() < full_len");
        log_err_p(LOG, "BPQBlock::initialise:  buf_len() = %zu",
            block->writable_contents()->buf_len());

        log_debug_p(LOG, "BPQBlock::initialise: reserving space in buffer %zu",
            full_len);

        block->writable_contents()->reserve(full_len);
        log_debug_p(LOG, "BPQBlock::initialise: new buf_len() = %zu",
            block->writable_contents()->buf_len());
    }

    buf = block->data();


    // BPQ Kind must be 0 or 1
    if ( *(block->data()) != 0 &&
         *(block->data()) != 1 ) {
        log_err_p(LOG, "BPQBlock::initialise: block->data() = %c (should be 0|1)",
            *(block->data()));
        return BP_FAIL;
    }

    /**************************************************************************
     * Begin extracting block info
     *************************************************************************/

    // BPQ-kind             1-byte
    if ( i < length ) {
        log_debug_p(LOG, "BPQBlock::initialise: extracting kind");
        kind_ = (kind_t) buf[i++];
        log_debug_p(LOG, "BPQBlock::initialise: kind = %d", kind_);
    } else {
    	log_err_p(LOG, "Error decoding BPQ kind");
    	return BP_FAIL;
    }

    // matching rule type   1-byte
    if ( i < length ) {
        matching_rule_ = (u_int) buf[i++];
        log_debug_p(LOG, "BPQBlock::initialise: matching rule = %u", matching_rule_);
    } else {
    	log_err_p(LOG, "Error decoding BPQ matching rule");
    	return BP_FAIL;
    }

    // query-len			SDNV
    if ( i < length &&
        (decoding_len = SDNV::decode(&(buf[i]), length - i, &query_len_)) >= 0 ) {
        i += decoding_len;
        log_debug_p(LOG, "BPQBlock::initialise: query len = %u", query_len_);
    } else {
        log_err_p(LOG, "Error decoding BPQ query length");
        return BP_FAIL;
    }

    // query-value           n-bytes
    if ( (i+query_len_) < length ) {
        query_val_ = (u_char*) malloc ( sizeof(u_char) * query_len_ );

        for (j=0; query_val_ != NULL && i < length && j < query_len_; i++, j++)
            query_val_[j] = buf[i];

        log_debug_p(LOG, "BPQBlock::initialise: query val = %s", query_val_);

    } else {
        query_val_ = NULL;
        log_err_p(LOG, "Error extracting BPQ query value");
        return BP_FAIL;
    }

    if ( i < length &&
        (decoding_len = SDNV::decode(&(buf[i]), length - i, &frag_count)) >= 0 ) {
        i += decoding_len;
        log_debug_p(LOG, "BPQBlock::initialise: frag count = %u", frag_count);
    } else {
        log_err_p(LOG, "Error decoding BPQ fragment count");
        return BP_FAIL;
    }


	for (j=0;  i < length && j < frag_count; j++) {

		if ( (decoding_len = SDNV::decode(&(buf[i]), length - i, &frag_off)) >= 0 ) {
			i += decoding_len;
			log_debug_p(LOG, "BPQBlock::initialise: frag offset = %u", frag_off);
		} else {
			log_err_p(LOG, "Error decoding BPQ fragment offset");
			return BP_FAIL;
		}

		if ( (decoding_len = SDNV::decode(&(buf[i]), length - i, &frag_len)) >= 0 ) {
			i += decoding_len;
			log_debug_p(LOG, "BPQBlock::initialise: frag length = %u", frag_len);
		} else {
			log_err_p(LOG, "Error decoding BPQ fragment length");
			return BP_FAIL;
		}


		BPQFragment frag(frag_off, frag_len);
		add_fragment(frag);
	}



    return BP_SUCCESS;
}

} // namespace dtn

